/*
 * The contents of this file are subject to the Mozilla Public License Version 1.1
 * (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at <http://www.mozilla.org/MPL/>.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, WITHOUT
 * WARRANTY OF ANY KIND, either express or implied. See the License for the specific
 * language governing rights and limitations under the License.
 * 
 * The Original Code is the Venice Web Community System.
 * 
 * The Initial Developer of the Original Code is Eric J. Bowersox <erbo@silcom.com>,
 * for Silverwrist Design Studios.  Portions created by Eric J. Bowersox are
 * Copyright (C) 2001 Eric J. Bowersox/Silverwrist Design Studios.  All Rights Reserved.
 * 
 * Contributor(s): 
 */
package com.silverwrist.venice.core.impl;

import java.io.*;
import java.sql.*;
import java.util.*;
import org.apache.log4j.*;
import com.silverwrist.util.StringUtil;
import com.silverwrist.venice.db.*;
import com.silverwrist.venice.security.AuditRecord;
import com.silverwrist.venice.core.*;

class TopicMessageUserContextImpl implements TopicMessageContext
{
  /*--------------------------------------------------------------------------------
   * Static data members
   *--------------------------------------------------------------------------------
   */

  private static Category logger = Category.getInstance(TopicMessageUserContextImpl.class.getName());

  private static final int MAX_ATTACH = 1048576;  // TODO: should be a configurable parameter

  /*--------------------------------------------------------------------------------
   * Attributes
   *--------------------------------------------------------------------------------
   */

  private EngineBackend engine;
  private ConferenceBackend conf;
  private DataPool datapool;
  private long postid;
  private long parent;
  private int num;
  private int linecount;
  private int creator_uid;
  private java.util.Date posted;
  private boolean hidden;
  private int scribble_uid;
  private java.util.Date scribble_date;
  private String pseud;
  private int datalen;
  private String filename;
  private String mimetype;
  private boolean nuked = false;
  private String creator_cache = null;
  private String text_cache = null;

  /*--------------------------------------------------------------------------------
   * Constructors
   *--------------------------------------------------------------------------------
   */

  protected TopicMessageUserContextImpl(EngineBackend engine, ConferenceBackend conf, DataPool datapool,
					long postid, long parent, int num, int linecount, int creator_uid,
					java.util.Date posted, boolean hidden, int scribble_uid,
					java.util.Date scribble_date, String pseud, int datalen,
					String filename, String mimetype)
  {
    this.engine = engine;
    this.conf = conf;
    this.datapool = datapool;
    this.postid = postid;
    this.parent = parent;
    this.num = num;
    this.linecount = linecount;
    this.creator_uid = creator_uid;
    this.posted = posted;
    this.hidden = hidden;
    this.scribble_uid = scribble_uid;
    this.scribble_date = scribble_date;
    this.pseud = pseud;
    this.datalen = datalen;
    this.filename = filename;
    this.mimetype = mimetype;

  } // end constructor

  TopicMessageUserContextImpl(EngineBackend engine, ConferenceBackend conf, DataPool datapool,
			      long postid, long parent, int num, int linecount, int creator_uid,
			      java.util.Date posted, String pseud)
  {
    this.engine = engine;
    this.conf = conf;
    this.datapool = datapool;
    this.postid = postid;
    this.parent = parent;
    this.num = num;
    this.linecount = linecount;
    this.creator_uid = creator_uid;
    this.posted = posted;
    this.hidden = false;
    this.scribble_uid = 0;
    this.scribble_date = null;
    this.pseud = pseud;
    this.datalen = 0;
    this.filename = null;
    this.mimetype = null;

  } // end constructor

  /*--------------------------------------------------------------------------------
   * Internal functions
   *--------------------------------------------------------------------------------
   */

  private static String quickGetUserName(Connection conn, int uid) throws SQLException
  {
    Statement stmt = conn.createStatement();
    ResultSet rs = stmt.executeQuery("SELECT username FROM users WHERE uid = " + String.valueOf(uid) + ";");
    if (rs.next())
      return rs.getString(1);
    else
      return "(unknown)";

  } // end quickGetUserName

  private void refresh(Connection conn) throws SQLException
  {
    Statement stmt = conn.createStatement();
    StringBuffer sql = new StringBuffer("SELECT posts.hidden, posts.scribble_uid, posts.scribble_date, "
					+ "posts.pseud, postattach.datalen, postattach.filename, "
					+ "postattach.mimetype FROM posts LEFT JOIN postattach ON "
					+ "posts.postid = postattach.postid WHERE posts.postid = ");
    sql.append(postid).append(';');
    ResultSet rs = stmt.executeQuery(sql.toString());
    if (rs.next())
    { // update a variety of fields
      hidden = rs.getBoolean(1);
      scribble_uid = rs.getInt(2);
      scribble_date = SQLUtil.getFullDateTime(rs,3);
      pseud = rs.getString(4);
      datalen = rs.getInt(5);
      filename = rs.getString(6);
      mimetype = rs.getString(7);

    } // end if
    else
    { // the post has been nuked - update accordingly
      linecount = 0;
      creator_uid = -1;
      posted = null;
      hidden = false;
      scribble_uid = -1;
      scribble_date = null;
      pseud = null;
      datalen = 0;
      filename = null;
      mimetype = null;
      nuked = true;
      creator_cache = null;
      text_cache = null;

    } // end else

  } // end refresh

  /*--------------------------------------------------------------------------------
   * Implementations from interface TopicMessageContext
   *--------------------------------------------------------------------------------
   */

  public long getPostID()
  {
    return postid;

  } // end getPostID

  public long getParentPostID()
  {
    return parent;

  } // end getParentPostID

  public int getPostNumber()
  {
    return num;

  } // end getPostNumber

  public int getNumLines()
  {
    return linecount;

  } // end getNumLines

  public int getCreatorUID()
  {
    return creator_uid;

  } // end getCreatorUID

  public String getCreatorName() throws DataException
  {
    if (creator_cache==null)
    { // we don't have the user name yet, get it out of the database
      if (nuked)
	return null;  // post nuked!
      Connection conn = null;

      try
      { // use a database connection to get the user name
	conn = datapool.getConnection();
	refresh(conn);
	if (nuked)
	  return null;  // post nuked!
	creator_cache = quickGetUserName(conn,creator_uid);

      } // end try
      catch (SQLException e)
      { // turn this into a DataException
	logger.error("DB error reading user name: " + e.getMessage(),e);
	throw new DataException("unable to retrieve user name: " + e.getMessage(),e);

      } // end catch
      finally
      { // make sure we release the connection before we go
	if (conn!=null)
	  datapool.releaseConnection(conn);

      } // end finally

    } // end if

    return creator_cache;

  } // end getCreatorName

  public java.util.Date getPostDate()
  {
    return posted;

  } // end getPostDate

  public boolean isHidden()
  {
    return hidden;

  } // end isHidden

  public boolean isScribbled()
  {
    return (scribble_date!=null);

  } // end isScribbled

  public boolean isNuked()
  {
    return nuked;

  } // end isNuked

  public java.util.Date getScribbleDate()
  {
    return scribble_date;

  } // end getScribbleDate

  public String getPseud()
  {
    return pseud;

  } // return pseud

  public String getBodyText() throws DataException
  {
    if (text_cache==null)
    { // we don't have the body text yet, go get it
      Connection conn = null;
      if (nuked)
	return null;  // post nuked!

      try
      { // use a database connection to get the body text
	conn = datapool.getConnection();
	refresh(conn);

	if (nuked)
	  return null;  // post nuked!

	if (scribble_date==null)
	{ // let's go get the body text!
	  Statement stmt = conn.createStatement();
	  ResultSet rs = stmt.executeQuery("SELECT data FROM postdata WHERE postid = "
					   + String.valueOf(postid) + ";");
	  if (rs.next())
	    text_cache = rs.getString(1);
	  else
	    return "Data Missing";  // FUTURE: throw an exception?

	} // end if
	else // for scribbled posts, we return the scribbler's name only
	  text_cache = quickGetUserName(conn,scribble_uid);

      } // end try
      catch (SQLException e)
      { // turn this into a DataException
	logger.error("DB error reading post data: " + e.getMessage(),e);
	throw new DataException("unable to retrieve post data: " + e.getMessage(),e);

      } // end catch
      finally
      { // make sure we release the connection before we go
	if (conn!=null)
	  datapool.releaseConnection(conn);

      } // end finally

    } // end if

    return text_cache;

  } // end getBodyText

  public boolean hasAttachment()
  {
    return (mimetype!=null);

  } // end hasAttachment

  public String getAttachmentType()
  {
    return mimetype;

  } // end getAttachmentType

  public String getAttachmentFilename()
  {
    return ((mimetype!=null) ? filename : null);

  } // end getAttachmentFilename

  public int getAttachmentLength()
  {
    return ((mimetype!=null) ? datalen : 0);

  } // end getAttachmentLength

  public InputStream getAttachmentData() throws AccessError, DataException
  {
    if (nuked || (scribble_date!=null))
    { // this would be an exercise in futility!
      logger.error("cannot attach to a nuked or scribbled message");
      throw new AccessError("You cannot attach data to a message that no longer exists.");

    } // end if

    Connection conn = null;
    InputStream rc = null;
    try
    { // open up a database connection
      conn = datapool.getConnection();

      // make sure we have current data
      refresh(conn);
      if (nuked || (scribble_date!=null))
      { // this would be an exercise in futility!
	logger.error("cannot attach to a nuked or scribbled message");
	throw new AccessError("You cannot attach data to a message that no longer exists.");

      } // end if

      if (mimetype==null)
      { // there is no attachment data!
	logger.error("no attachment data to get");
	throw new AccessError("There is no attachment data for this message.");

      } // end if

      // Create the statement and the SQL we need to retrieve the attachment.
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("SELECT data FROM postattach WHERE postid = ");
      sql.append(postid).append(';');

      // Execute the query!
      ResultSet rs = stmt.executeQuery(sql.toString());
      if (!(rs.next()))
      { // there is no attachment data!
	logger.error("no attachment data to get");
	throw new AccessError("There is no attachment data for this message.");

      } // end if

      // Since the InputStream we get from JDBC will probably go away when the connection is dropped, we
      // need to make a temporary copy of it, to a ByteArrayOutputStream.  (Attachments can only be
      // 1 Mb in size, so this shouldn't be a big problem.)
      InputStream sqldata = rs.getBinaryStream(1);
      ByteArrayOutputStream copy = new ByteArrayOutputStream(datalen);
      byte[] buffer = new byte[4096];
      int rd = sqldata.read(buffer);
      while (rd>=0)
      { // write, then read again
	if (rd>0)
	  copy.write(buffer,0,rd);
	rd = sqldata.read(buffer);

      } // end while

      // Close both our streams, making sure we create the stream we need for the return value.
      sqldata.close();
      rc = new ByteArrayInputStream(copy.toByteArray());
      copy.close();

    } // end try
    catch (SQLException e)
    { // turn this into a DataException
      logger.error("DB error retrieving attachment: " + e.getMessage(),e);
      throw new DataException("unable to retrieve attachment data: " + e.getMessage(),e);

    } // end catch
    catch (IOException e)
    { // turn this into a DataException too
      logger.error("I/O error copying attachment data: " + e.getMessage(),e);
      throw new DataException("unable to retrieve attachment data: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

    return rc;

  } // end getAttachmentData

  public boolean canHide()
  {
    return (((creator_uid==conf.realUID()) && (!conf.userIsAnonymous())) || conf.userCanHide());

  } // end canHide

  public boolean canScribble()
  {
    return (((creator_uid==conf.realUID()) && (!conf.userIsAnonymous())) || conf.userCanScribble());

  } // end canScribble

  public boolean canNuke()
  {
    return conf.userCanNuke();

  } // end canNuke

  public void setHidden(boolean flag) throws DataException, AccessError
  {
    if (conf.userIsAnonymous())
      return;  // no-op
    if ((creator_uid!=conf.realUID()) && !(conf.userCanHide()))
    { // we can't change the hidden status!
      logger.error("trying to set hidden status of post w/o permission!");
      throw new AccessError("You are not permitted to change the hidden status of this message.");

    } // end if

    if (nuked || (scribble_date!=null) || (hidden==flag))
      return;  // changing the status of a nuked or scribbled post is futile

    Connection conn = null;
    AuditRecord ar = null;

    try
    { // open up a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();

      // lock the tables we reference
      stmt.executeUpdate("LOCK TABLES posts WRITE, postattach READ;");
      try
      { // first, make sure we have the right status for our post
	refresh(conn);
	if (nuked || (scribble_date!=null))
	  return;  // changing the status of a nuked or scribbled post is futile
	if (hidden==flag)
	  return;  // this is a no-op

	// update the "hidden" flag in the database
	StringBuffer sql = new StringBuffer("UPDATE posts SET hidden = ");
	sql.append(flag ? '1' : '0').append(" WHERE postid = ").append(postid).append(';');
	stmt.executeUpdate(sql.toString());

	hidden = flag;  // store flag

      } // end try
      finally
      { // unlock the tables before we go
	Statement ulk_stmt = conn.createStatement();
	ulk_stmt.executeUpdate("UNLOCK TABLES;");

      } // end finally

      // record what we did in an audit record
      ar = new AuditRecord(AuditRecord.HIDE_MESSAGE,conf.realUID(),conf.userRemoteAddress(),
			   conf.realSIGID(),"conf=" + String.valueOf(conf.realConfID()) + ",post="
			   + String.valueOf(postid),flag ? "hide" : "unhide");
      
    } // end try
    catch (SQLException e)
    { // turn this into a DataException
      logger.error("DB error setting hidden status: " + e.getMessage(),e);
      throw new DataException("unable to set hidden status: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      try
      { // save off the audit record before we go, though
	if ((ar!=null) && (conn!=null))
	  ar.store(conn);

      } // end try
      catch (SQLException e)
      { // we couldn't store the audit record!
	logger.error("DB error saving audit record: " + e.getMessage(),e);

      } // end catch

      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end setHidden

  public void scribble() throws DataException, AccessError
  {
    if (conf.userIsAnonymous())
      return;  // no-op
    if ((creator_uid!=conf.realUID()) && !(conf.userCanScribble()))
    { // we can't scribble this post
      logger.error("trying to scribble post w/o permission!");
      throw new AccessError("You are not permitted to scribble this message.");

    } // end if

    if (nuked || (scribble_date!=null))
      return;  // scribbling a nuked or scribbled post is futile

    Connection conn = null;
    AuditRecord ar = null;

    try
    { // open up a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();

      // lock the tables we reference
      stmt.executeUpdate("LOCK TABLES posts WRITE, postdata WRITE, postattach WRITE;");
      try
      { // first, make sure we have the right status for our post
	refresh(conn);
	if (nuked || (scribble_date!=null))
	  return;  // scribbling a nuked or scribbled post is futile

	// First, set the appropriate "scribbled" information in the "header".
	StringBuffer sql = new StringBuffer("UPDATE posts SET linecount = 0, hidden = 0, scribble_uid = ");
	sql.append(conf.realUID()).append(", scribble_date = '");
	java.util.Date now = new java.util.Date();
	final String scribble_pseud = "<EM><B>(Scribbled)</B></EM>";  // TODO: configurable option
	sql.append(SQLUtil.encodeDate(now)).append("', pseud = '").append(scribble_pseud);
	sql.append("' WHERE postid = ").append(postid).append(';');
	if (logger.isDebugEnabled())
	  logger.debug("SQL: " + sql.toString());
	stmt.executeUpdate(sql.toString());

	// Determine if we need to "rub out" the post before we delete it.
	ResultSet rs = stmt.executeQuery("SELECT LENGTH(data) FROM postdata WHERE postid = "
					 + String.valueOf(postid) + ";");
	if (rs.next())
	{ // use this data to overwrite the post with X's
	  int len = rs.getInt(1);
	  if (len>0)
	  { // construct the "rubout" statement and execute it
	    sql.setLength(0);
	    sql.append("UPDATE postdata SET data = '");
	    while (len>0)
	    { // generate a string of X's the length of the post
	      sql.append('X');
	      len--;

	    } // end while

	    sql.append("' WHERE postid = ").append(postid).append(';');
	    stmt.executeUpdate(sql.toString());

	  } // end if
	  // else not much need to do a rubout

	} // end if
	// else don't try...we're deleting the row anyway

	// Delete the actual post data row.
	sql.setLength(0);
	sql.append("DELETE FROM postdata WHERE postid = ").append(postid).append(';');
	stmt.executeUpdate(sql.toString());

	// Delete the attachment data row.
	// FUTURE: can we do an overwrite on the attachment the way we did on the post data?
	sql.setLength(0);
	sql.append("DELETE FROM postattach WHERE postid = ").append(postid).append(';');
	stmt.executeUpdate(sql.toString());

	// Update our internal data fields.
	linecount = 0;
	hidden = false;
	scribble_uid = conf.realUID();
	scribble_date = now;
	pseud = scribble_pseud;
	text_cache = null;
	
      } // end try
      finally
      { // unlock the tables before we go
	Statement ulk_stmt = conn.createStatement();
	ulk_stmt.executeUpdate("UNLOCK TABLES;");

      } // end finally

      // record what we did in an audit record
      ar = new AuditRecord(AuditRecord.SCRIBBLE_MESSAGE,conf.realUID(),conf.userRemoteAddress(),
			   conf.realSIGID(),"conf=" + String.valueOf(conf.realConfID()) + ",post="
			   + String.valueOf(postid));
      
    } // end try
    catch (SQLException e)
    { // turn this into a DataException
      logger.error("DB error scribbling post: " + e.getMessage(),e);
      throw new DataException("unable to scribble message: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      try
      { // save off the audit record before we go, though
	if ((ar!=null) && (conn!=null))
	  ar.store(conn);

      } // end try
      catch (SQLException e)
      { // we couldn't store the audit record!
	logger.error("DB error saving audit record: " + e.getMessage(),e);

      } // end catch

      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end scribble

  public void nuke() throws DataException, AccessError
  {
    if (!(conf.userCanNuke()))
    { // we can't scribble this post
      logger.error("trying to nuke post w/o permission!");
      throw new AccessError("You are not permitted to nuke this message.");

    } // end if

    if (nuked)
      return;  // nuking a nuked post is futile

    Connection conn = null;
    AuditRecord ar = null;

    try
    { // open up a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();

      // lock the tables we reference
      stmt.executeUpdate("LOCK TABLES posts WRITE, postdata WRITE, postattach WRITE, postdogear WRITE;");

      try
      { // first, make sure we have the right status for our post
	refresh(conn);
	if (nuked)
	  return;  // nuking a nuked post is futile

	// Delete any and all references to this post!
	stmt.executeUpdate("DELETE FROM posts WHERE postid = " + String.valueOf(postid) + ";");
	stmt.executeUpdate("DELETE FROM postdata WHERE postid = " + String.valueOf(postid) + ";");
	stmt.executeUpdate("DELETE FROM postattach WHERE postid = " + String.valueOf(postid) + ";");
	stmt.executeUpdate("DELETE FROM postdogear WHERE postid = " + String.valueOf(postid) + ";");

	// Update our internal variables.
	linecount = 0;
	creator_uid = -1;
	posted = null;
	hidden = false;
	scribble_uid = -1;
	scribble_date = null;
	pseud = null;
	datalen = 0;
	filename = null;
	mimetype = null;
	nuked = true;
	creator_cache = null;
	text_cache = null;

      } // end try
      finally
      { // unlock the tables before we go
	Statement ulk_stmt = conn.createStatement();
	ulk_stmt.executeUpdate("UNLOCK TABLES;");

      } // end finally

      // record what we did in an audit record
      ar = new AuditRecord(AuditRecord.NUKE_MESSAGE,conf.realUID(),conf.userRemoteAddress(),
			   conf.realSIGID(),"conf=" + String.valueOf(conf.realConfID()) + ",post="
			   + String.valueOf(postid));

    } // end try
    catch (SQLException e)
    { // turn this into a DataException
      logger.error("DB error nuking post: " + e.getMessage(),e);
      throw new DataException("unable to nuke message: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      try
      { // save off the audit record before we go, though
	if ((ar!=null) && (conn!=null))
	  ar.store(conn);

      } // end try
      catch (SQLException e)
      { // we couldn't store the audit record!
	logger.error("DB error saving audit record: " + e.getMessage(),e);

      } // end catch

      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end nuke

  public void attachData(String m_type, String file, int length, InputStream data)
      throws AccessError, DataException
  {
    if (mimetype!=null)
    { // the message already has an attachment
      logger.error("tried to attach data to a message that already has it!");
      throw new AccessError("This message already has an attachment.");

    } // end if

    if (creator_uid!=conf.realUID())
    { // you can't attach to this message!
      logger.error("tried to attach data to a message that's not yours!");
      throw new AccessError("You are not permitted to add an attachment to this message.");

    } // end if

    if (nuked || (scribble_date!=null))
    { // this would be an exercise in futility!
      logger.error("cannot attach to a nuked or scribbled message");
      throw new AccessError("You cannot attach data to a message that no longer exists.");

    } // end if

    if (StringUtil.isStringEmpty(m_type))
    { // no MIME type specified
      logger.error("no MIME type specified for attachment");
      throw new AccessError("MIME type of attachment data not specified.");

    } // end if

    if (StringUtil.isStringEmpty(file))
    { // no MIME type specified
      logger.error("no filename specified for attachment");
      throw new AccessError("Filename of attachment data not specified.");

    } // end if

    if (length<=0)
    { // a length of 0 or less is just WRONG
      logger.error("non-positive length specified for attachment");
      throw new AccessError("Invalid attachment length.");

    } // end if
    else if (length>MAX_ATTACH)
    { // the attachment is too damn long!
      logger.error("attachment is too long (" + String.valueOf(length) + " bytes)");
      throw new AccessError("The attachment is too long to store.  Maximum available length is "
			    + String.valueOf(MAX_ATTACH) + " bytes.");

    } // end else if

    Connection conn = null;
    AuditRecord ar = null;

    try
    { // open up a database connection
      conn = datapool.getConnection();

      // make sure we have the right status to upload
      refresh(conn);
      if (nuked || (scribble_date!=null))
      { // this would be an exercise in futility!
	logger.error("cannot attach to a nuked or scribbled message");
	throw new AccessError("You cannot attach data to a message that no longer exists.");

      } // end if

      // Build the SQL statement that inserts the attachment.  Note the use of the "?" to specify the
      // BLOB parameter (the attachment data itself); this comes later.
      StringBuffer sql =
          new StringBuffer("INSERT INTO postattach (postid, datalen, filename, mimetype, data) VALUES (");
      sql.append(postid).append(", ").append(length).append(", '").append(SQLUtil.encodeString(file));
      sql.append("', '").append(SQLUtil.encodeString(m_type)).append("', ?);");

      // Prepare the statement, set the BLOB parameter, and execute it.
      PreparedStatement stmt = conn.prepareStatement(sql.toString());
      stmt.setBinaryStream(1,data,length);
      stmt.executeUpdate();

      // Save off the local attachment values.
      datalen = length;
      filename = file;
      mimetype = m_type;

      // Generate an audit record indicating what we did.
      ar = new AuditRecord(AuditRecord.UPLOAD_ATTACHMENT,conf.realUID(),conf.userRemoteAddress(),
			   conf.realSIGID(),"conf=" + String.valueOf(conf.realConfID()) + ",post="
			   + String.valueOf(postid),"len=" + String.valueOf(length) + ",type=" + m_type
			   + ",name=" + file);

    } // end try
    catch (SQLException e)
    { // turn this into a DataException
      logger.error("DB error saving attachment: " + e.getMessage(),e);
      throw new DataException("unable to save attachment data: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      try
      { // save off the audit record before we go, though
	if ((ar!=null) && (conn!=null))
	  ar.store(conn);

      } // end try
      catch (SQLException e)
      { // we couldn't store the audit record!
	logger.error("DB error saving audit record: " + e.getMessage(),e);

      } // end catch

      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end attachData

  /*--------------------------------------------------------------------------------
   * External static operations
   *--------------------------------------------------------------------------------
   */

  static List loadMessageRange(EngineBackend engine, ConferenceBackend conf, DataPool datapool, int topicid,
			       int post_low, int post_high) throws DataException
  {
    if (logger.isDebugEnabled())
      logger.debug("loadMessageRange for conf # " + String.valueOf(conf.realConfID()) + ", topic #"
		   + String.valueOf(topicid) + ", range [" + String.valueOf(post_low) + ", "
		   + String.valueOf(post_high) + "]");

    Vector rc = new Vector();
    Connection conn = null;      // pooled database connection

    try
    { // get a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();

      // run a query to get all the posts in a particular topic
      StringBuffer sql =
	  new StringBuffer("SELECT p.postid, p.parent, p.num, p.linecount, p.creator_uid, p.posted, "
			   + "p.hidden, p.scribble_uid,	p.scribble_date, p.pseud, a.datalen, a.filename, "
			   + "a.mimetype FROM posts p LEFT JOIN postattach a ON p.postid = a.postid "
			   + "WHERE p.topicid = ");
      sql.append(topicid).append(" AND p.num >= ").append(post_low).append(" AND p.num <= ");
      sql.append(post_high).append(" ORDER BY p.num ASC;");
      if (logger.isDebugEnabled())
	logger.debug("SQL: " + sql.toString());
      ResultSet rs = stmt.executeQuery(sql.toString());

      while (rs.next())
      { // create implementation objects and shove them into the return vector
	TopicMessageContext val =
	  new TopicMessageUserContextImpl(engine,conf,datapool,rs.getLong(1),rs.getLong(2),rs.getInt(3),
					  rs.getInt(4),rs.getInt(5),SQLUtil.getFullDateTime(rs,6),
					  rs.getBoolean(7),rs.getInt(8),SQLUtil.getFullDateTime(rs,9),
					  rs.getString(10),rs.getInt(11),rs.getString(12),rs.getString(13));
	rc.add(val);

      } // end while

    } // end try
    catch (SQLException e)
    { // turn SQLException into data exception
      logger.error("DB error reading message entries: " + e.getMessage(),e);
      throw new DataException("unable to retrieve messages: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

    return new ReadOnlyVector(rc);  // wrap the return vector

  } // end loadMessageRange

  static TopicMessageContext loadMessage(EngineBackend engine, ConferenceBackend conf, DataPool datapool,
					 int topicid, int message_num) throws DataException
  {
    if (logger.isDebugEnabled())
      logger.debug("loadMessage for conf # " + String.valueOf(conf.realConfID()) + ", topic #"
		   + String.valueOf(topicid) + ", message " + String.valueOf(message_num));

    Connection conn = null;      // pooled database connection

    try
    { // get a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();

      // run a query to get all the posts in a particular topic
      StringBuffer sql =
	  new StringBuffer("SELECT p.postid, p.parent, p.num, p.linecount, p.creator_uid, p.posted, "
			   + "p.hidden, p.scribble_uid,	p.scribble_date, p.pseud, a.datalen, a.filename, "
			   + "a.mimetype FROM posts p LEFT JOIN postattach a ON p.postid = a.postid "
			   + "WHERE p.topicid = ");
      sql.append(topicid).append(" AND p.num = ").append(message_num).append(';');
      if (logger.isDebugEnabled())
	logger.debug("SQL: " + sql.toString());
      ResultSet rs = stmt.executeQuery(sql.toString());

      if (rs.next())  // create an object reference and return it
	return new TopicMessageUserContextImpl(engine,conf,datapool,rs.getLong(1),rs.getLong(2),rs.getInt(3),
		                               rs.getInt(4),rs.getInt(5),SQLUtil.getFullDateTime(rs,6),
					       rs.getBoolean(7),rs.getInt(8),SQLUtil.getFullDateTime(rs,9),
					       rs.getString(10),rs.getInt(11),rs.getString(12),
					       rs.getString(13));

      // indicates an error...
      throw new DataException("Message not found.");

    } // end try
    catch (SQLException e)
    { // turn SQLException into data exception
      logger.error("DB error reading message entry: " + e.getMessage(),e);
      throw new DataException("unable to retrieve message: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end loadMessage

  static TopicMessageContext getMessage(EngineBackend engine, ConferenceBackend conf, DataPool datapool,
					long postid) throws DataException
  {
    if (logger.isDebugEnabled())
      logger.debug("getMessage for conf # " + String.valueOf(conf.realConfID()) + ", post #"
		   + String.valueOf(postid));

    Connection conn = null;      // pooled database connection

    try
    { // get a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();

      StringBuffer sql =
	  new StringBuffer("SELECT p.postid, p.parent, p.num, p.linecount, p.creator_uid, p.posted, "
			   + "p.hidden, p.scribble_uid,	p.scribble_date, p.pseud, a.datalen, a.filename, "
			   + "a.mimetype FROM topics t, posts p LEFT JOIN postattach a ON p.postid = a.postid "
			   + "WHERE t.topicid = p.topicid AND t.confid = ");
      sql.append(conf.realConfID()).append(" AND p.postid = ").append(postid).append(';');
      if (logger.isDebugEnabled())
	logger.debug("SQL: " + sql.toString());
      ResultSet rs = stmt.executeQuery(sql.toString());

      if (rs.next())  // create an object reference and return it
	return new TopicMessageUserContextImpl(engine,conf,datapool,rs.getLong(1),rs.getLong(2),rs.getInt(3),
		                               rs.getInt(4),rs.getInt(5),SQLUtil.getFullDateTime(rs,6),
					       rs.getBoolean(7),rs.getInt(8),SQLUtil.getFullDateTime(rs,9),
					       rs.getString(10),rs.getInt(11),rs.getString(12),
					       rs.getString(13));

      // indicates an error...
      throw new DataException("Message not found.");

    } // end try    
    catch (SQLException e)
    { // turn SQLException into data exception
      logger.error("DB error reading message entries: " + e.getMessage(),e);
      throw new DataException("unable to retrieve messages: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end getMessage

} // end class TopicMessageUserContextImpl
