/*
 * The contents of this file are subject to the Mozilla Public License Version 1.1
 * (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at <http://www.mozilla.org/MPL/>.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, WITHOUT
 * WARRANTY OF ANY KIND, either express or implied. See the License for the specific
 * language governing rights and limitations under the License.
 * 
 * The Original Code is the Venice Web Communities System.
 * 
 * The Initial Developer of the Original Code is Eric J. Bowersox <erbo@silcom.com>,
 * for Silverwrist Design Studios.  Portions created by Eric J. Bowersox are
 * Copyright (C) 2001 Eric J. Bowersox/Silverwrist Design Studios.  All Rights Reserved.
 * 
 * Contributor(s): 
 */
package com.silverwrist.venice.core.impl;

import java.sql.*;
import java.util.*;
import org.apache.log4j.*;
import org.w3c.dom.*;
import com.silverwrist.util.StringUtil;
import com.silverwrist.util.DOMElementHelper;
import com.silverwrist.venice.core.*;
import com.silverwrist.venice.db.*;
import com.silverwrist.venice.htmlcheck.*;
import com.silverwrist.venice.htmlcheck.dict.*;
import com.silverwrist.venice.htmlcheck.filters.*;
import com.silverwrist.venice.security.AuditRecord;
import com.silverwrist.venice.security.PasswordHash;
import com.silverwrist.venice.security.DefaultLevels;

public class VeniceEngineImpl implements VeniceEngine, EngineBackend
{
  /*--------------------------------------------------------------------------------
   * Internal class implementing the Country interface
   *--------------------------------------------------------------------------------
   */

  static class CountryImpl implements Country
  {
    private String code;
    private String name;

    public CountryImpl(String code, String name)
    {
      this.code = code;
      this.name = name;

    } // end constructor

    public String getCode()
    {
      return code;

    } // end getCode

    public String getName()
    {
      return name;

    } // end getName

  } // end class CountryImpl

  /*--------------------------------------------------------------------------------
   * Internal class implementing the Language interface
   *--------------------------------------------------------------------------------
   */

  static class LanguageImpl implements Language
  {
    private String code;
    private String name;

    public LanguageImpl(String code, String name)
    {
      this.code = code;
      this.name = name;

    } // end constructor

    public String getCode()
    {
      return code;

    } // end getCode

    public String getName()
    {
      return name;

    } // end getName

  } // end class LanguageImpl

  /*--------------------------------------------------------------------------------
   * Internal class storing feature information.
   *--------------------------------------------------------------------------------
   */

  class VeniceFeatureDef implements SIGFeature
  {
    int code;
    String symbol;
    String name;
    String applet;
    boolean is_default;
    boolean is_locked;
    boolean is_hidden;
    boolean requires_read;
    int sequence;
    int min_level;
    
    VeniceFeatureDef(int code, ResultSet rs) throws SQLException
    {
      this.code = code;
      this.is_default = rs.getBoolean("is_default");
      this.is_locked = rs.getBoolean("is_locked");
      this.is_hidden = rs.getBoolean("is_hidden");
      this.requires_read = rs.getBoolean("require_read");
      this.sequence = rs.getInt("sequence");
      this.min_level = rs.getInt("min_level");
      this.symbol = rs.getString("symbol");
      this.name = rs.getString("name");
      this.applet = rs.getString("applet");

    } // end constructor

    public int getFeatureCode()
    {
      return code;

    } // end getFeatureCode

    public String getSymbol()
    {
      return symbol;

    } // end getSymbol

    public String getName()
    {
      return name;

    } // end getName

    public String getApplet()
    {
      return applet;

    } // end getApplet

    public int getSequence()
    {
      return sequence;

    } // end getSequence

    boolean featureAllowed(int level, boolean has_read)
    {
      if (level<min_level)
	return false;
      if (requires_read)
	return has_read;
      else
	return true;

    } // end featureAllowed

    boolean isVisible()
    {
      return !is_hidden;

    } // end isVisible

    boolean isLocked()
    {
      return is_locked;

    } // end isLocked

    boolean isDefault()
    {
      return is_default;

    } // end isDefault

  } // end class VeniceFeatureDef

  /*--------------------------------------------------------------------------------
   * Static data values
   *--------------------------------------------------------------------------------
   */

  private static Category logger = Category.getInstance(VeniceEngineImpl.class.getName());

  /*--------------------------------------------------------------------------------
   * Attributes
   *--------------------------------------------------------------------------------
   */

  private Document config = null;                     // configuration data
  private DataPool datapool = null;                   // database connection pool
  private Random rng;                                 // random number generator
  private Properties email_props = null;              // email properties
  private javax.mail.Session mailsession = null;      // email session object
  private Hashtable stock_messages = null;            // stock messages holder
  private Hashtable sig_objects = new Hashtable();    // holder for SIGCoreData objects
  private VeniceFeatureDef[] features;                // master feature table
  private Hashtable feature_syms = new Hashtable();   // hashtable mapping symbols to features
  private Hashtable conf_objects = new Hashtable();   // holder for ConferenceCoreData objects
  private HTMLCheckerConfig[] html_configs;           // holder for HTML checker configurations
  private int[] gp_ints;                              // global integer parameters

  /*--------------------------------------------------------------------------------
   * Constructor
   *--------------------------------------------------------------------------------
   */

  public VeniceEngineImpl()
  {
    // Initialize the random number generator.
    java.util.Date seed_date = new java.util.Date();
    rng = new Random(seed_date.getTime());

  } // end constructor

  /*--------------------------------------------------------------------------------
   * Internal functions
   *--------------------------------------------------------------------------------
   */

  private void checkInitialized()
  {
    if (config==null)
      throw new InternalStateError("Venice engine not initialized!");

  } // end checkInitialized

  private void loadDefaults(Statement stmt) throws SQLException, DataException
  {
    final String query =
        "SELECT posts_per_page, old_posts_at_top, max_search_page, max_sig_mbr_page, max_conf_mbr_page "
      + "FROM globals;";
    ResultSet rs = stmt.executeQuery(query);
    if (!(rs.next()))
      throw new DataException("Globals table does not appear to be loaded!");

    // stash the globals from the database
    gp_ints[IP_POSTSPERPAGE] = rs.getInt(1);
    gp_ints[IP_POSTSATTOP] = rs.getInt(2);
    gp_ints[IP_MAXSEARCHRETURN] = rs.getInt(3);
    gp_ints[IP_MAXSIGMEMBERDISPLAY] = rs.getInt(4);
    gp_ints[IP_MAXCONFMEMBERDISPLAY] = rs.getInt(5);
    
  } // end loadDefaults

  /*--------------------------------------------------------------------------------
   * Implementations from interface VeniceEngine
   *--------------------------------------------------------------------------------
   */

  /**
   * Loads the Venice engine configuration from the named XML file and creates appropriate subobjects,
   * such as the data pool.
   *
   * @param config The XML document conating the configuration information for the engine.
   * @exception com.silverwrist.venice.core.ConfigException An error occurred during engine configuration;
   *                                                        the engine could not be initialized.
   * @exception com.silverwrist.venice.core.DataException The Venice engine data could not be loaded.
   */
  public void initialize(Document config) throws ConfigException, DataException
  {
    if (this.config!=null)
    { // already configured!
      logger.error("Venice engine already initialized");
      throw new ConfigException("Venice engine already initialized");

    } // end if

    this.config = config;

    Vector dictionary_tmp;
    try
    { // first, verify that this is a valid configuration
      Element root = config.getDocumentElement();
      if (!(root.getTagName().equals("venice-config")))
      { // not the correct root tag name
	logger.fatal("config document is not a <venice-config/> document (root tag: <"
		     + root.getTagName() + "/>)");
	throw new ConfigException("document is not a <venice-config/> document",root);

      } // end if

      // Get the <database/> section.
      DOMElementHelper root_h = new DOMElementHelper(root);
      Element db_sect = root_h.getSubElement("database");
      if (db_sect==null)
      { // unable to find the database section
	logger.fatal("config document has no <database/> section");
	throw new ConfigException("no <database/> section found in config file",root);

      } // end if

      try
      { // allocate the data pool object
	datapool = new DataPool(db_sect);

      } // end try
      catch (SQLException e)
      { // could not initialize the SQL database!  this is a big fat error!
	logger.fatal("data pool init failed - " + e.getMessage(),e);
	throw new ConfigException("database access initialization failed - " + e.getMessage(),e);

      } // end catch

      Element email_sect = root_h.getSubElement("email");
      if (email_sect==null)
      { // unable to find the database section
	logger.fatal("config document has no <email/> section");
	throw new ConfigException("no <email/> section found in config file",root);

      } // end if

      // initialize the email properties and get a mail session object
      email_props = new Properties();
      email_props.put("mail.transport.protocol","smtp");
      DOMElementHelper email_sect_h = new DOMElementHelper(email_sect);
      email_props.put("mail.smtp.host",email_sect_h.getSubElementText("smtp-host"));
      email_props.put("mail.from",email_sect_h.getSubElementText("mail-from-addr"));
      email_props.put("com.silverwrist.venice.email.mailer",email_sect_h.getSubElementText("mailer"));
      mailsession = javax.mail.Session.getInstance(email_props,null);

      Element dict_sect = root_h.getSubElement("dictionary");
      if (dict_sect==null)
      { // unable to find the database section
	logger.fatal("config document has no <dictionary/> section");
	throw new ConfigException("no <dictionary/> section found in config file",root);

      } // end if

      // Retrieve the list of dictionary files to load into the spellchecker.
      dictionary_tmp = new Vector();
      NodeList dict_nodes = dict_sect.getChildNodes();
      int j;
      for (j=0; j<dict_nodes.getLength(); j++)
      { // scan the <dictionary> element looking for <file> elements
	Node dn = dict_nodes.item(j);
	if ((dn.getNodeType()==Node.ELEMENT_NODE) && (dn.getNodeName().equals("file")))
	{ // store the file name as a vector element
	  Element del = (Element)dn;
	  DOMElementHelper h = new DOMElementHelper(del);
	  dictionary_tmp.add(h.getElementText());

	} // end if

      } // end for
      
      Element msg_sect = root_h.getSubElement("messages");
      if (msg_sect==null)
      { // unable to find the database section
	logger.fatal("config document has no <messages/> section");
	throw new ConfigException("no <messages/> section found in config file",root);

      } // end if

      // Initialize the stock messages list.
      stock_messages = new Hashtable();
      NodeList msg_nodes = msg_sect.getChildNodes();
      for (j=0; j<msg_nodes.getLength(); j++)
      { // examine all subnodes to add them to the message text
	Node msgn = msg_nodes.item(j);
	if (msgn.getNodeType()==Node.ELEMENT_NODE)
	{ // add it to the hash table by its tag name
	  Element msgel = (Element)msgn;
	  DOMElementHelper h = new DOMElementHelper(msgel);
	  String txt = h.getElementText();
	  if (txt!=null)
	    stock_messages.put(msgel.getTagName(),txt.trim());

	} // end if

      } // end for

      if (logger.isDebugEnabled())
	logger.debug(stock_messages.size() + " stock messages loaded from config");

    } // end try
    catch (ConfigException ce)
    { // before we leave on a ConfigException, nuke the important data
      this.config = null;
      datapool = null;
      throw ce;

    } // end catch

    // Allocate the global parameter arrays.
    gp_ints = new int[5];

    // initialize anything that requires us to pull from the database
    Connection conn = null;
    try
    { // get a connection from the data pool
      conn = datapool.getConnection();

      // load the master feature table
      Statement stmt = conn.createStatement();
      ResultSet rs = stmt.executeQuery("SELECT * FROM refsigftr ORDER BY ftr_code;");
      VeniceFeatureDef[] tmp_array = new VeniceFeatureDef[64];
      int max_value = -1;
      while (rs.next())
      { // store the values as we get them
	int ndx = rs.getInt("ftr_code");
	tmp_array[ndx] = new VeniceFeatureDef(ndx,rs);
	if (ndx>max_value)
	  max_value = ndx;

      } // end while

      // store the real feature table as an array
      features = new VeniceFeatureDef[++max_value];
      System.arraycopy(tmp_array,0,features,0,max_value);
      if (logger.isDebugEnabled())
	logger.debug(max_value + " features loaded from database");

      // load the global defaults
      loadDefaults(stmt);

    } // end try
    catch (SQLException e)
    { // convert exceptions to DataException
      logger.fatal("Database error loading engine data: " + e.getMessage(),e);
      throw new DataException("error initializing Venice engine: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

    for (int i=0; i<features.length; i++) // insert feature symbols into hashtable
      feature_syms.put(features[i].getSymbol(),features[i]);

    // Here is where we create the HTML Checker and all the goodies it relies on.
    // Start by creating some of the subsidiary objects that get added to HTML Checker configs.
    EmailRewriter email_rewriter = new EmailRewriter();
    HTMLEncodingFilter html_filter = new HTMLEncodingFilter();
    SQLEncodingFilter sql_filter = new SQLEncodingFilter();
    SpellingRewriter spell_rewriter = new SpellingRewriter();
    URLRewriter url_rewriter = new URLRewriter();
    PostLinkRewriter postlink_rewriter = new PostLinkRewriter(datapool);
    UserNameRewriter username_rewriter = new UserNameRewriter(datapool);

    // Create the LazyLexicon that holds our dictionary files, and add it to the SpellingRewriter.
    String[] dictfiles = new String[dictionary_tmp.size()];
    for (int i=0; i<dictionary_tmp.size(); i++)
      dictfiles[i] = (String)(dictionary_tmp.get(i));
    LazyTreeLexicon lex = new LazyTreeLexicon(dictfiles);
    spell_rewriter.addDictionary(lex);

    html_configs = new HTMLCheckerConfig[4];  // create the array

    // Create the HTML checker config used to post body text to the database.
    HTMLCheckerConfig cfg = HTMLCheckerCreator.create();
    cfg.setWordWrapLength((short)55);
    cfg.setProcessAngles(true);
    cfg.setProcessParens(true);
    cfg.setDiscardHTMLTags(false);
    cfg.addOutputFilter(html_filter);
    cfg.addOutputFilter(sql_filter);
    cfg.addRawOutputFilter(sql_filter);
    cfg.addStringRewriter(email_rewriter);
    cfg.addStringRewriter(url_rewriter);
    cfg.addTagRewriter(postlink_rewriter);
    cfg.addTagRewriter(username_rewriter);
    cfg.addTagRewriter(email_rewriter);
    cfg.addTagRewriter(url_rewriter);
    cfg.addParenRewriter(username_rewriter);
    cfg.configureNormalTagSet();
    html_configs[HTMLC_POST_BODY] = cfg;

    // Create the HTML checker config used to post pseuds and topic names to the database.
    cfg = HTMLCheckerCreator.create();
    cfg.setWordWrapLength((short)0);
    cfg.setProcessAngles(true);
    cfg.setProcessParens(false);
    cfg.setDiscardHTMLTags(false);
    cfg.addOutputFilter(html_filter);
    cfg.addOutputFilter(sql_filter);
    cfg.addRawOutputFilter(sql_filter);
    cfg.configureRestrictedTagSet();
    html_configs[HTMLC_POST_PSEUD] = cfg;

    // Create the HTML checker config used to preview body text.
    cfg = HTMLCheckerCreator.create();
    cfg.setWordWrapLength((short)55);
    cfg.setProcessAngles(true);
    cfg.setProcessParens(true);
    cfg.setDiscardHTMLTags(false);
    cfg.addOutputFilter(html_filter);
    cfg.addStringRewriter(email_rewriter);
    cfg.addStringRewriter(url_rewriter);
    cfg.addWordRewriter(spell_rewriter);
    cfg.addTagRewriter(postlink_rewriter);
    cfg.addTagRewriter(username_rewriter);
    cfg.addTagRewriter(email_rewriter);
    cfg.addTagRewriter(url_rewriter);
    cfg.addParenRewriter(username_rewriter);
    cfg.configureNormalTagSet();
    html_configs[HTMLC_PREVIEW_BODY] = cfg;
    
    // Create the HTML checker config used to escape body text and pseuds for putting back in the post box.
    cfg = HTMLCheckerCreator.create();
    cfg.setWordWrapLength((short)0);
    cfg.setProcessAngles(false);
    cfg.setProcessParens(false);
    cfg.setDiscardHTMLTags(false);
    cfg.addOutputFilter(html_filter);
    html_configs[HTMLC_ESCAPE_BODY_PSEUD] = cfg;

    if (logger.isDebugEnabled())
      logger.debug("initialize() complete :-)");

  } // end initialize

  public int getNumFeatures()
  {
    checkInitialized();
    return features.length;

  } // end getNumFeatures

  public BitSet getAllFeaturesMask()
  {
    checkInitialized();
    BitSet rc = new BitSet();
    for (int i=0; i<features.length; i++)
      if (features[i].isVisible())
	rc.set(i);

    return rc;

  } // end getAllFeaturesMask

  public List getCountryList() throws DataException
  {
    checkInitialized();
    Connection conn = null;
    Vector rc = new Vector();

    try
    { // do a SELECT on the refcountry table to load the master country list
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      ResultSet rs = stmt.executeQuery("SELECT code, name FROM refcountry ORDER BY name;");

      while (rs.next())
      { // load up a stack of memory objects with the country list
	Country c = new CountryImpl(rs.getString(1),rs.getString(2));
	rc.add(c);

      } // end while

      if (logger.isDebugEnabled())
	logger.debug("getCountryList(): loaded " + rc.size() + " country names");

    } // end try
    catch (SQLException e)
    { // remap to a DataException
      logger.error("DB error loading countries: " + e.getMessage(),e);
      throw new DataException("unable to load country data: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

    return new ReadOnlyVector(rc);

  } // end getCountryList

  public List getLanguageList() throws DataException
  {
    checkInitialized();
    Connection conn = null;
    Vector rc = new Vector();

    try
    { // do a SELECT on the refcountry table to load the master country list
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      ResultSet rs = stmt.executeQuery("SELECT code, name FROM reflanguage ORDER BY name;");

      while (rs.next())
      { // load up a stack of memory objects with the country list
	Language l = new LanguageImpl(rs.getString(1),rs.getString(2));
	rc.add(l);

      } // end while

      if (logger.isDebugEnabled())
	logger.debug("getLanguageList(): loaded " + rc.size() + " language names");

    } // end try
    catch (SQLException e)
    { // remap to a DataException
      logger.error("DB error loading languages: " + e.getMessage(),e);
      throw new DataException("unable to load language data: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

    return new ReadOnlyVector(rc);

  } // end getLanguageList

  public String getNameOfCountry(String code)
  {
    checkInitialized();
    return getCountryNameForCode(code);

  } // end getNameOfCountry

  public UserContext createUserContext(String remote_addr) throws DataException
  {
    checkInitialized();
    UserContextImpl uci = new UserContextImpl(this,datapool);
    uci.loadAnonymous(remote_addr);
    if (logger.isDebugEnabled())
      logger.debug("createUserContext(): context loaded :-)");
    return (UserContext)uci;

  } // end createUserContext

  public String getEmailAddressForUser(String username) throws DataException, AccessError
  {
    checkInitialized();
    Connection conn = null;
      
    try
    { // look for a user name matching this user record
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      ResultSet rs = stmt.executeQuery("SELECT c.email FROM users u, contacts c WHERE u.contactid = "
				       + "c.contactid AND u.username = '" + SQLUtil.encodeString(username)
				       + "';");
      if (!(rs.next()))
      { // could not find email address
	logger.warn("getEmailAddressForUser(\"" + username + "\") => not found!");
	throw new AccessError("Unable to find email address for user \"" + username +
			      "\".  Please contact the system administrator for assistance.");

      } // end if

      String rc = rs.getString(1);
      if (logger.isDebugEnabled())
	logger.debug("getEmailAddressForUser(\"" + username + "\") => \"" + rc + "\"");
      return rc;

    } // end try
    catch (SQLException e)
    { // database error - this is a DataException
      logger.error("DB error getting email address for user: " + e.getMessage(),e);
      throw new DataException("unable to access user data: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end getEmailAddressForUser

  public void sendPasswordReminder(String username) throws DataException, AccessError, EmailException
  {
    checkInitialized();
    Connection conn = null;
      
    try
    { // look for a user name matching this user record
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("SELECT c.email, u.passreminder FROM users u, contacts c WHERE "
					  + "u.contactid = c.contactid AND u.username = '");
      sql.append(SQLUtil.encodeString(username)).append("';");
      ResultSet rs = stmt.executeQuery(sql.toString());
      if (!(rs.next()))
      { // the email address could not be found
	logger.error("can't find email to send password reminder to \"" + username + "\"");
	throw new AccessError("Unable to find email address for user \"" + username +
			      "\".  Please contact the system administrator for assistance.");

      } // end if

      String email_addr = rs.getString(1);
      if (logger.isDebugEnabled())
	logger.debug("sendPasswordReminder(\"" + username + "\") going out to \"" + email_addr + "\"");

      // Create the message to be sent.
      String message = getStockMessage("reminder");
      if (message==null)
      { // what? the message does not exist?
	logger.error("\"reminder\" message does not exist in config file!");
	throw new DataException("INTERNAL: reminder message not specified");

      } // end if

      message = StringUtil.replaceAllInstances(message,"$USERNAME",username);
      message = StringUtil.replaceAllInstances(message,"$REMINDER",rs.getString(2));

      // Create the emailer and send the message.
      SimpleEmailer emailer = createEmailer();
      emailer.setTo(email_addr);
      emailer.setSubject("Venice Password Reminder Message");
      emailer.setText(message);
      emailer.send();
      if (logger.isDebugEnabled())
	logger.debug("...email sent");

    } // end try
    catch (SQLException e)
    { // database error - this is a DataException
      logger.error("DB error getting email address/reminder for user: " + e.getMessage(),e);
      throw new DataException("unable to access user data: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end sendPasswordReminder

  public UserContext createNewAccount(String remote_addr, String username, String password, String reminder)
      throws DataException, AccessError
  {
    checkInitialized();
    Connection conn = null;
    AuditRecord ar = null;
    String encode_username = SQLUtil.encodeString(username);
    // email confirmation # is between 1000000 and 9999999
    int confirm_num = getNewConfirmationNumber();
    int new_uid;
    java.util.Date created = null;

    if (logger.isDebugEnabled())
      logger.debug("createNewAccount(\"" + username + "\",<password>,<reminder>)...");
      
    try
    { // look to see if the user name is already present
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      stmt.executeUpdate("LOCK TABLES users WRITE, userprefs WRITE, sigmember WRITE, topconfig WRITE;");
      try
      { // make sure the user name isn't there already
	ResultSet rs = stmt.executeQuery("SELECT uid FROM users WHERE username = '" + encode_username + "';");
	if (rs.next())
	{ // the user account already exists
	  logger.warn("username \"" + username + "\" already exists");
	  throw new AccessError("That user name already exists.  Please try again.");

	} // end if

	// perform the insert of the user's new data
	PasswordHash phash = new PasswordHash(password);
	StringBuffer sql = new StringBuffer("INSERT INTO users (username, passhash, email_confnum, "
					    + "base_lvl, created, lastaccess, passreminder) VALUES ('");
	sql.append(encode_username).append("', '").append(phash.toString()).append("', ");
	sql.append(confirm_num).append(", ").append(DefaultLevels.newUser()).append(", '");
	created = new java.util.Date();
	sql.append(SQLUtil.encodeDate(created)).append("', '").append(SQLUtil.encodeDate(created));
	sql.append("', ").append(SQLUtil.encodeStringArg(reminder)).append(");");
	stmt.executeUpdate(sql.toString());

	// what is the new user ID?
	rs = stmt.executeQuery("SELECT LAST_INSERT_ID();");
	if (!(rs.next()))
	{ // the readback failed...
	  logger.error("readback of new user's UID failed");
	  throw new DataException("unable to read back new user ID");

	} // end if

	new_uid = rs.getInt(1);
	if (logger.isDebugEnabled())
	  logger.debug("...created user \"" + username + "\" with UID " + new_uid);

	// add a UserPrefs record for this user, too
	sql.setLength(0);
	sql.append("INSERT INTO userprefs (uid) VALUES (").append(new_uid).append(");");
	stmt.executeUpdate(sql.toString());

	if (logger.isDebugEnabled())
	  logger.debug("...created userprefs");

	// get the list of SIG IDs the anonymous user is a member of
	rs = stmt.executeQuery("SELECT sigmember.sigid, sigmember.locked FROM users, sigmember "
			       + "WHERE sigmember.uid = users.uid AND users.is_anon = 1;");
	sql.setLength(0);
	while (rs.next())
	{ // set up to insert into the sigmember table
	  if (sql.length()==0)
	    sql.append("INSERT INTO sigmember (sigid, uid, granted_lvl, locked) VALUES ");
	  else
	    sql.append(", ");
	  sql.append("(").append(rs.getInt(1)).append(", ").append(new_uid).append(", ");
	  sql.append(DefaultLevels.memberSIG()).append(", ").append(rs.getInt(2)).append(")");
				 
	} // end while

	if (sql.length()>0)
	{ // execute the big update
	  sql.append(';');
	  stmt.executeUpdate(sql.toString());

	} // end if

	if (logger.isDebugEnabled())
	  logger.debug("...loaded default SIG memberships");

	// get the "top" configuration for this user
	rs = stmt.executeQuery("SELECT topconfig.row, topconfig.col, topconfig.partid, topconfig.param "
			       + "FROM topconfig, users WHERE topconfig.uid = users.uid AND "
			       + "users.is_anon = 1;");
	sql.setLength(0);
	while (rs.next())
	{ // set up to insert into the topconfig table
	  if (sql.length()==0)
	    sql.append("INSERT INTO topconfig (uid, row, col, partid, param) VALUES ");
	  else
	    sql.append(", ");
	  sql.append("(").append(new_uid).append(", ").append(rs.getInt(1)).append(", ");
	  sql.append(rs.getInt(2)).append(", ").append(SQLUtil.encodeStringArg(rs.getString(3)));
	  sql.append(", ").append(SQLUtil.encodeStringArg(rs.getString(4))).append(')');

	} // end while

	if (sql.length()>0)
	{ // execute the big update
	  sql.append(';');
	  stmt.executeUpdate(sql.toString());

	} // end if

	if (logger.isDebugEnabled())
	  logger.debug("...loaded default top view config");

      } // end try
      finally
      { // make sure the tables get unlocked before we go
	Statement ulk_stmt = conn.createStatement();
	ulk_stmt.executeUpdate("UNLOCK TABLES;");

      } // end finally

      // the operation was a success - give back an audit record
      ar = new AuditRecord(AuditRecord.ACCOUNT_CREATE,new_uid,remote_addr);

    } // end try
    catch (SQLException e)
    { // database error - this is a DataException
      logger.error("DB error creating user: " + e.getMessage(),e);
      throw new DataException("unable to create user record: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      try
      { // save off the audit record before we go, though
	if ((ar!=null) && (conn!=null))
	  ar.store(conn);

      } // end try
      catch (SQLException e)
      { // we couldn't store the audit record!
	logger.error("DB error saving audit record: " + e.getMessage(),e);

      } // end catch

      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

    // create a new context for the user (they're now effectively logged in)
    UserContextImpl rc = new UserContextImpl(this,datapool);
    rc.loadNewUser(remote_addr,new_uid,DefaultLevels.newUser(),username,confirm_num,created,created);
    if (logger.isDebugEnabled())
      logger.debug("...created new user context");
    return rc;

  } // end createNewAccount

  public boolean aliasExists(String alias, int exist_sigid)
  {
    checkInitialized();
    Connection conn = null;

    try
    { // do a SELECT on the sigs table
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("SELECT sigid FROM sigs WHERE alias = '");
      sql.append(alias).append("'");
      if (exist_sigid>0)
	sql.append(" AND sigid <> ").append(exist_sigid);
      sql.append(';');
      ResultSet rs = stmt.executeQuery(sql.toString());
      return rs.next();

    } // end try
    catch (SQLException e)
    { // force to return null
      logger.warn("DB error looking up SIG: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

    return false;  // this is a "default" return

  } // end aliasExists

  public boolean isValidCategoryID(int catid)
  {
    checkInitialized();
    if (catid==-1)
      return true;  // valid by definition, it means "Top"
    Connection conn = null;

    try
    { // do a SELECT on the sigs table
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("SELECT catid FROM refcategory WHERE catid = ");
      sql.append(catid).append(';');
      ResultSet rs = stmt.executeQuery(sql.toString());
      return rs.next();

    } // end try
    catch (SQLException e)
    { // force to return null
      logger.warn("DB error looking up category: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

    return false;  // this is a "default" return

  } // end isValidCategoryID

  public List searchForUsers(int field, int mode, String term, int offset, int count) throws DataException
  {
    if (logger.isDebugEnabled())
      logger.debug("User search: field = " + field + ", mode = " + mode + ", term '" + term + "', offset = "
		   + offset + ", count = " + count);

    Vector rc = new Vector();    // return from this function
    Connection conn = null;      // pooled database connection

    try
    { // get a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("SELECT u.uid, u.username, u.description, c.given_name, "
					  + "c.family_name, c.locality, c.region, c.country FROM users u, "
					  + "contacts c WHERE u.contactid = c.contactid AND ");

      switch (field)
      { // determine which field to search on
	case SearchMode.FIELD_USER_NAME:
	  sql.append("u.username ");
	  break;

	case SearchMode.FIELD_USER_DESCRIPTION:
	  sql.append("u.description ");
	  break;

	case SearchMode.FIELD_USER_GIVEN_NAME:
	  sql.append("c.given_name ");
	  break;

	case SearchMode.FIELD_USER_FAMILY_NAME:
	  sql.append("c.family_name ");
	  break;

	default:
	  throw new DataException("invalid search field specified");

      } // end switch

      switch (mode)
      { // compose SQL in different ways depending on the search term
	case SearchMode.SEARCH_PREFIX:
	  sql.append("LIKE '").append(SQLUtil.encodeStringWildcards(term)).append("%'");
	  break;

	case SearchMode.SEARCH_SUBSTRING:
	  sql.append("LIKE '%").append(SQLUtil.encodeStringWildcards(term)).append("%'");
	  break;

	case SearchMode.SEARCH_REGEXP:
	  sql.append("REGEXP '").append(SQLUtil.encodeString(term)).append('\'');
	  break;

	default:
	  throw new DataException("invalid search mode specified");

      } // end switch

      sql.append(" AND u.is_anon = 0 ORDER BY u.username LIMIT ").append(offset).append(", ");
      sql.append(count+1).append(';');

      if (logger.isDebugEnabled())
	logger.debug("SQL: " + sql.toString());

      // launch the search!
      ResultSet rs = stmt.executeQuery(sql.toString());

      while (rs.next())
      { // add all the found users to the list
	UserFoundImpl ufi = new UserFoundImpl(rs.getInt(1),rs.getString(2),rs.getString(3),rs.getString(4),
					      rs.getString(5),rs.getString(6),rs.getString(7),rs.getString(8));
	UserFound tmp = (UserFound)ufi;
	rc.add(tmp);

      } // end while

    } // end try
    catch (SQLException e)
    { // turn SQLException into data exception
      logger.error("DB error reading user entries: " + e.getMessage(),e);
      throw new DataException("unable to retrieve user information: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

    return new ReadOnlyVector(rc);

  } // end searchForUsers

  public int getSearchUserCount(int field, int mode, String term) throws DataException
  {
    if (logger.isDebugEnabled())
      logger.debug("User search: field = " + field + ", mode = " + mode + ", term '" + term + "'");

    Connection conn = null;      // pooled database connection

    try
    { // get a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("SELECT COUNT(*) FROM users u, contacts c WHERE u.contactid = "
					  + "c.contactid AND ");

      switch (field)
      { // determine which field to search on
	case SearchMode.FIELD_USER_NAME:
	  sql.append("u.username ");
	  break;

	case SearchMode.FIELD_USER_DESCRIPTION:
	  sql.append("u.description ");
	  break;

	case SearchMode.FIELD_USER_GIVEN_NAME:
	  sql.append("c.given_name ");
	  break;

	case SearchMode.FIELD_USER_FAMILY_NAME:
	  sql.append("c.family_name ");
	  break;

	default:
	  throw new DataException("invalid search field specified");

      } // end switch

      switch (mode)
      { // compose SQL in different ways depending on the search term
	case SearchMode.SEARCH_PREFIX:
	  sql.append("LIKE '").append(SQLUtil.encodeStringWildcards(term)).append("%'");
	  break;

	case SearchMode.SEARCH_SUBSTRING:
	  sql.append("LIKE '%").append(SQLUtil.encodeStringWildcards(term)).append("%'");
	  break;

	case SearchMode.SEARCH_REGEXP:
	  sql.append("REGEXP '").append(SQLUtil.encodeString(term)).append('\'');
	  break;

	default:
	  throw new DataException("invalid search mode specified");

      } // end switch

      sql.append(" AND u.is_anon = 0;");


      if (logger.isDebugEnabled())
	logger.debug("SQL: " + sql.toString());

      // launch the search!
      ResultSet rs = stmt.executeQuery(sql.toString());

      if (!(rs.next()))
	throw new InternalStateError("getSearchUserCount search failure (MUST have 1 row!)");

      return rs.getInt(1);

    } // end try
    catch (SQLException e)
    { // turn SQLException into data exception
      logger.error("DB error reading user entries: " + e.getMessage(),e);
      throw new DataException("unable to retrieve user information: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end getSearchUserCount

  public int getStdNumSearchResults()
  {
    return gp_ints[IP_MAXSEARCHRETURN];

  } // end getStdNumSearchResults

  public boolean isEmailAddressBanned(String email)
  {
    Connection conn = null;      // pooled database connection

    try
    { // get a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();

      // create and execute the right SQL statement
      StringBuffer sql = new StringBuffer("SELECT by_uid FROM emailban WHERE address = '");
      sql.append(SQLUtil.encodeString(email)).append("';");
      ResultSet rs = stmt.executeQuery(sql.toString());
      return rs.next();

    } // end try
    catch (SQLException e)
    { // if the database is horked, return false
      logger.warn("DB error looking email address: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

    return false;

  } // end isEmailAddressBanned

  public boolean confAliasExists(String alias)
  {
    checkInitialized();
    Connection conn = null;

    try
    { // do a SELECT on the confalias table
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("SELECT confid FROM confalias WHERE alias = '");
      sql.append(alias).append("';");
      ResultSet rs = stmt.executeQuery(sql.toString());
      return rs.next();

    } // end try
    catch (SQLException e)
    { // force to return false
      logger.warn("DB error looking up conference: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

    return false;  // this is a "default" return

  } // end confAliasExists

  public HTMLChecker getEscapingChecker()
  {
    return html_configs[HTMLC_ESCAPE_BODY_PSEUD].createHTMLChecker();

  } // end getEscapingChecker

  public int getNumPostsPerPage()
  {
    return gp_ints[IP_POSTSPERPAGE];

  } // end getNumPostsPerPage

  public int getNumOldPostsBeforeNew()
  {
    return gp_ints[IP_POSTSATTOP];

  } // end getNumOldPostsBeforeNew

  public int getMaxNumConferenceMembersDisplay()
  {
    return gp_ints[IP_MAXCONFMEMBERDISPLAY];

  } // end getMaxNumConferenceMembersDisplay

  public int getMaxNumSIGMembersDisplay()
  {
    return gp_ints[IP_MAXSIGMEMBERDISPLAY];

  } // end getMaxNumSIGMembersDisplay

  /*--------------------------------------------------------------------------------
   * Implementations from interface EngineBackend
   *--------------------------------------------------------------------------------
   */

  public SimpleEmailer createEmailer()
  {
    checkInitialized();
    return new SimpleEmailer(email_props,mailsession);

  } // end createEmailer

  public String getStockMessage(String key)
  {
    checkInitialized();
    return (String)(stock_messages.get(key));

  } // end getStockMessage

  public int getNewConfirmationNumber()
  {
    checkInitialized();
    return rng.nextInt(9000000) + 1000000;

  } // end getNewConfirmationNumber

  public String getCountryNameForCode(String code)
  {
    checkInitialized();
    if (code==null)
      return null;
    Connection conn = null;

    try
    { // do a SELECT on the refcountry table
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      ResultSet rs = stmt.executeQuery("SELECT name FROM refcountry WHERE code = '" + code + "';");
      if (rs.next())
      { // found it!
	String rc = rs.getString("name");
	if (logger.isDebugEnabled())
	  logger.debug("getCountryNameForCode(\"" + code + "\") => \"" + rc + "\"");
	return rc;

      } // end if

    } // end try
    catch (SQLException e)
    { // force to return null
      logger.warn("DB error looking up country code: " + e.getMessage(),e);
      if (logger.isDebugEnabled())
	logger.debug("getCountryNameForCode(\"" + code + "\") => not found (data error)");
      return null;

    } // end catch
    finally
    { // make sure the connection is released before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

    if (logger.isDebugEnabled())
      logger.debug("getCountryNameForCode(\"" + code + "\") => not found");
    return null;  // if get here, it wasn't found

  } // end getCountryNameForCode

  public String getLanguageNameForCode(String code)
  {
    checkInitialized();
    if (code==null)
      return null;
    Connection conn = null;

    try
    { // do a SELECT on the refcountry table
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      ResultSet rs = stmt.executeQuery("SELECT name FROM reflanguage WHERE CODE = '" + code + "';");
      if (rs.next())
      { // found it!
	String rc = rs.getString("name");
	if (logger.isDebugEnabled())
	  logger.debug("getLanguageNameForCode(\"" + code + "\") => \"" + rc + "\"");
	return rc;

      } // end if

    } // end try
    catch (SQLException e)
    { // force to return null
      logger.warn("DB error looking up language code: " + e.getMessage(),e);
      if (logger.isDebugEnabled())
	logger.debug("getLanguageNameForCode(\"" + code + "\") => not found (data error)");
      return null;

    } // end catch
    finally
    { // make sure the connection is released before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

    if (logger.isDebugEnabled())
      logger.debug("getLanguageNameForCode(\"" + code + "\") => not found");
    return null;  // if get here, it wasn't found

  } // end getLanguageNameForCode

  public synchronized SIGData getSIGDataObject(int sigid) throws DataException
  {
    checkInitialized();
    if (logger.isDebugEnabled())
      logger.debug("getSIGDataObject(" + sigid + ")...");

    Integer the_sigid = new Integer(sigid);
    SIGData sd = (SIGData)(sig_objects.get(the_sigid));
    if (sd==null)
    { // create a new SIGCoreData object and save it off
      sd = new SIGCoreData(this,datapool,sigid);
      sig_objects.put(the_sigid,sd);
      if (logger.isDebugEnabled())
	logger.debug("...created new object");

    } // end if
    else
    {  // this is an extra reference to the SIGDataObject
      sd.rd_addRef();
      if (logger.isDebugEnabled())
	logger.debug("...using cached object");

    } // end else

    return sd;

  } // end getSIGDataObject

  public synchronized void detachSIGDataObject(int sigid)
  {
    checkInitialized();
    if (logger.isDebugEnabled())
      logger.debug("detachSIGDataObject(" + sigid + ")...");

    Integer the_sigid = new Integer(sigid);
    SIGData sd = (SIGData)(sig_objects.get(the_sigid));
    if (sd!=null)
    { // pull the reference out of our hashtable
      sig_objects.remove(the_sigid);
      sd.rd_release();
      if (logger.isDebugEnabled())
	logger.debug("...reference detached");
      
    } // end if

  } // end detachSIGDataObject

  public BitSet getLockedFeaturesMask()
  {
    checkInitialized();
    BitSet rc = new BitSet();
    for (int i=0; i<features.length; i++)
      if (features[i].isLocked())
	rc.set(i);

    return rc;

  } // end getLockedFeaturesMask

  public List getSIGFeatureSet(BitSet enabled_features, int level, boolean read_privs)
  {
    checkInitialized();
    Vector rc = new Vector();
    for (int i=0; i<features.length; i++)
      if (enabled_features.get(i) && features[i].featureAllowed(level,read_privs))
      { // this feature must be included in the eventual output set
	SIGFeature elt = features[i];
	boolean insert_me = true;
	for (int j=0; insert_me && (j<rc.size()); j++)
	  if (elt.getSequence()<((SIGFeature)(rc.get(j))).getSequence())
	  { // put the item in the list
	    rc.add(j,elt);
	    insert_me = false;

	  } // end if and for

	if (insert_me)  // insert at end by default
	  rc.addElement(elt);

      } // end if and for

    if (logger.isDebugEnabled())
      logger.debug("getSIGFeatureSet() loaded " + rc.size() + " elements");

    return new ReadOnlyVector(rc);  // wrap the vector for return

  } // end getSIGFeatureSet

  public String getAppletForFeature(int code)
  {
    checkInitialized();
    if ((code<0) || (code>=features.length))
    { // the index is out of bounds!
      logger.error("getAppletForFeature(" + code + ") => out of range");
      throw new IndexOutOfBoundsException("invalid feature code");

    } // end if

    String rc = features[code].getApplet();
    if (logger.isDebugEnabled())
      logger.debug("getAppletForFeature(" + code + ") => \"" + rc + "\"");
    return rc;

  } // end getAppletForFeature

  public boolean isValidInitialFeatureIndex(short ndx)
  {
    if ((ndx<0) || (ndx>=features.length))
      return false;
    return features[ndx].isVisible();

  } // end isValidInitialFeatureIndex

  public BitSet getDefaultFeaturesMask()
  {
    checkInitialized();
    BitSet rc = new BitSet();
    for (int i=0; i<features.length; i++)
      if (features[i].isDefault())
	rc.set(i);

    return rc;

  } // end getDefaultFeaturesMask

  public synchronized void registerNewSIG(SIGData sig)
  {
    checkInitialized();
    if (logger.isDebugEnabled())
      logger.debug("registerNewSIG(" + sig.getID() + ")...");

    Integer the_sigid = new Integer(sig.getID());
    if (sig_objects.get(the_sigid)!=null)
      throw new InternalStateError("SIGID " + the_sigid + " already exists...but it CAN'T HAVE!");

    // throw an extra reference on it and dump it in the object cache
    sig.rd_addRef();
    sig_objects.put(the_sigid,sig);
    if (logger.isDebugEnabled())
      logger.debug("...registered new object");

  } // end registerNewSIG

  public int getFeatureIndexBySymbol(String symbol)
  {
    VeniceFeatureDef fd = (VeniceFeatureDef)(feature_syms.get(symbol));
    if (fd!=null)
      return fd.getFeatureCode();
    else
      return -1;

  } // end getFeatureIndexBySymbol

  public boolean canAccessFeature(String symbol, int level, boolean has_read)
  {
    VeniceFeatureDef fd = (VeniceFeatureDef)(feature_syms.get(symbol));
    if (fd!=null)
      return fd.featureAllowed(level,has_read);
    else
      return false;

  } // end canAccessFeature

  public synchronized ConferenceData getConferenceDataObject(int confid) throws DataException
  {
    checkInitialized();
    if (logger.isDebugEnabled())
      logger.debug("getConferenceDataObject(" + confid + ")...");

    Integer the_confid = new Integer(confid);
    ConferenceData cd = (ConferenceData)(conf_objects.get(the_confid));
    if (cd==null)
    { // create a new ConferenceCoreData object and save it off
      cd = new ConferenceCoreData(this,datapool,confid);
      conf_objects.put(the_confid,cd);
      if (logger.isDebugEnabled())
	logger.debug("...created new object");

    } // end if
    else
    {  // this is an extra reference to the ConferenceDataObject
      cd.rd_addRef();
      if (logger.isDebugEnabled())
	logger.debug("...using cached object");

    } // end else

    return cd;

  } // end getConferenceDataObject

  public synchronized void detachConferenceDataObject(int confid)
  {
    checkInitialized();
    if (logger.isDebugEnabled())
      logger.debug("detachConferenceDataObject(" + confid + ")...");

    Integer the_confid = new Integer(confid);
    ConferenceData cd = (ConferenceData)(conf_objects.get(the_confid));
    if (cd!=null)
    { // remove it from our hashtable
      conf_objects.remove(the_confid);
      cd.rd_release();
      if (logger.isDebugEnabled())
	logger.debug("...detached reference");

    } // end if

  } // end detachConferenceDataObject

  public void saveAuditRecord(AuditRecord ar)
  {
    checkInitialized();
    Connection conn = null;

    if (ar==null)
      return;  // don't store a null record

    try
    { // get a connection and use it to store the audit record
      conn = datapool.getConnection();
      ar.store(conn);

    } // end try
    catch (SQLException e)
    { // just log an error if we screwed up
      logger.error("DB error saving audit record: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end saveAuditRecord

  public synchronized void registerNewConference(ConferenceData conf)
  {
    checkInitialized();
    if (logger.isDebugEnabled())
      logger.debug("registerNewConference(" + conf.getID() + ")...");

    Integer the_confid = new Integer(conf.getID());
    if (conf_objects.get(the_confid)!=null)
      throw new InternalStateError("ConfID " + the_confid + " already exists...but it CAN'T HAVE!");

    // throw an extra reference on it and dump it in the object cache
    conf.rd_addRef();
    conf_objects.put(the_confid,conf);
    if (logger.isDebugEnabled())
      logger.debug("...registered new object");

  } // end registerNewConference

  public HTMLChecker createCheckerObject(int type)
  {
    if ((type<0) || (type>=html_configs.length))
      throw new IllegalArgumentException("invalid HTML checker type index");
    return html_configs[type].createHTMLChecker();

  } // end createCheckerObject

  public int getParamInt(int selector)
  {
    return gp_ints[selector];

  } // end getParamInt

  public void forceParamReload() throws DataException
  {
    Connection conn = null;  // data pooled connection

    try
    { // get a connection and use it to reload
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      loadDefaults(stmt);

    } // end try
    catch (SQLException e)
    { // just log an error if we screwed up
      logger.error("DB error loading parameters: " + e.getMessage(),e);
      throw new DataException("Database error reloading parameters: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end forceParamReload

} // end class VeniceEngineImpl
