/*
 * The contents of this file are subject to the Mozilla Public License Version 1.1
 * (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at <http://www.mozilla.org/MPL/>.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, WITHOUT
 * WARRANTY OF ANY KIND, either express or implied. See the License for the specific
 * language governing rights and limitations under the License.
 * 
 * The Original Code is the Venice Web Community System.
 * 
 * The Initial Developer of the Original Code is Eric J. Bowersox <erbo@silcom.com>,
 * for Silverwrist Design Studios.  Portions created by Eric J. Bowersox are
 * Copyright (C) 2001 Eric J. Bowersox/Silverwrist Design Studios.  All Rights Reserved.
 * 
 * Contributor(s): 
 */
package com.silverwrist.venice.security;

import java.sql.*;
import java.util.*;
import com.silverwrist.venice.db.SQLUtil;
import com.silverwrist.venice.core.InternalStateError;

public class AuditRecord implements Audit
{
  /*--------------------------------------------------------------------------------
   * Attributes
   *--------------------------------------------------------------------------------
   */

  private long record;             // audit record identifier
  private java.util.Date when;     // the date/time of this audit event
  private int type;                // the audit event type
  private int uid;                 // the user ID
  private int sigid;               // the SIG ID
  private String ip;               // the IP address of the user
  private String[] data;           // the data values associated with the record
  private String descr = null;     // audit record description

  /*--------------------------------------------------------------------------------
   * Constructors
   *--------------------------------------------------------------------------------
   */

  public AuditRecord(int type, int uid, String ip, int sigid, String data1, String data2, String data3,
		     String data4)
  {
    setBaseData(type,uid,ip);
    this.sigid = sigid;
    setData(data1,data2,data3,data4);

  } // end constructor

  public AuditRecord(int type, int uid, String ip, int sigid, String data1, String data2, String data3)
  {
    setBaseData(type,uid,ip);
    this.sigid = sigid;
    setData(data1,data2,data3,null);

  } // end constructor

  public AuditRecord(int type, int uid, String ip, int sigid, String data1, String data2)
  {
    setBaseData(type,uid,ip);
    this.sigid = sigid;
    setData(data1,data2,null,null);

  } // end constructor

  public AuditRecord(int type, int uid, String ip, int sigid, String data1)
  {
    setBaseData(type,uid,ip);
    this.sigid = sigid;
    setData(data1,null,null,null);

  } // end constructor

  public AuditRecord(int type, int uid, String ip, int sigid)
  {
    setBaseData(type,uid,ip);
    this.sigid = sigid;
    setData(null,null,null,null);

  } // end constructor

  public AuditRecord(int type, int uid, String ip, String data1, String data2, String data3, String data4)
  {
    setBaseData(type,uid,ip);
    this.sigid = 0;
    setData(data1,data2,data3,data4);

  } // end constructor

  public AuditRecord(int type, int uid, String ip, String data1, String data2, String data3)
  {
    setBaseData(type,uid,ip);
    this.sigid = 0;
    setData(data1,data2,data3,null);

  } // end constructor

  public AuditRecord(int type, int uid, String ip, String data1, String data2)
  {
    setBaseData(type,uid,ip);
    this.sigid = 0;
    setData(data1,data2,null,null);

  } // end constructor

  public AuditRecord(int type, int uid, String ip, String data1)
  {
    setBaseData(type,uid,ip);
    this.sigid = 0;
    setData(data1,null,null,null);

  } // end constructor

  public AuditRecord(int type, int uid, String ip)
  {
    setBaseData(type,uid,ip);
    this.sigid = 0;
    setData(null,null,null,null);

  } // end constructor

  /*--------------------------------------------------------------------------------
   * Internal functions
   *--------------------------------------------------------------------------------
   */

  private void setBaseData(int type, int uid, String ip)
  {
    this.record = 0;
    this.when = null;
    this.type = type;
    this.uid = uid;
    this.ip = ip;

  } // end setBaseData

  private void setData(String data1, String data2, String data3, String data4)
  {
    data = new String[4];
    data[0] = data1;
    data[1] = data2;
    data[2] = data3;
    data[3] = data4;

  } // end setData

  /*--------------------------------------------------------------------------------
   * External operations
   *--------------------------------------------------------------------------------
   */

  public long getRecord()
  {
    return record;

  } // end getRecord

  public java.util.Date getDateTime()
  {
    return when;

  } // end getDateTime

  public int getType()
  {
    return type;

  } // end getType

  public int getUID()
  {
    return uid;

  } // end getType

  public int getSIGID()
  {
    return sigid;

  } // end getType

  public String getIPAddress()
  {
    return ip;

  } // end getIPAddress

  public String getData(int ndx)
  {
    return data[ndx];

  } // end getData

  public String getDescription()
  {
    return descr;

  } // end getDescription

  public void store(Connection conn) throws SQLException
  {
    if (record!=0)
      throw new InternalStateError("audit record " + String.valueOf(record) + " already stored!");

    Statement stmt = conn.createStatement();
    stmt.executeUpdate("LOCK TABLES audit WRITE;");
    try
    { // attempt to insert a record into the audit table
      StringBuffer sql = new StringBuffer("INSERT INTO audit (on_date, event, uid, sigid, ip, data1, data2, "
					  + "data3, data4) VALUES ('");
      java.util.Date now = new java.util.Date();
      sql.append(SQLUtil.encodeDate(now)).append("', ").append(type).append(", ").append(uid).append(", ");
      sql.append(sigid).append(", '").append(SQLUtil.encodeString(ip)).append("', ");
      sql.append(SQLUtil.encodeStringArg(data[0])).append(", ").append(SQLUtil.encodeStringArg(data[1]));
      sql.append(", ").append(SQLUtil.encodeStringArg(data[2])).append(", ");
      sql.append(SQLUtil.encodeStringArg(data[3])).append(");");

      stmt.executeUpdate(sql.toString());

      ResultSet rs = stmt.executeQuery("SELECT LAST_INSERT_ID();");
      if (!(rs.next()))
	throw new InternalStateError("AuditRecord.store bogus query - must return at least 1 row!");
      record = rs.getLong(1);
      when = now;

    } // end try
    finally
    { // make sure we unlock the tables before we go
      Statement ulk_stmt = conn.createStatement();
      ulk_stmt.executeUpdate("UNLOCK TABLES;");

    } // end finally

  } // end store

} // end class AuditRecord
