/*
 * The contents of this file are subject to the Mozilla Public License Version 1.1
 * (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at <http://www.mozilla.org/MPL/>.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, WITHOUT
 * WARRANTY OF ANY KIND, either express or implied. See the License for the specific
 * language governing rights and limitations under the License.
 * 
 * The Original Code is the Venice Web Communities System.
 * 
 * The Initial Developer of the Original Code is Eric J. Bowersox <erbo@silcom.com>,
 * for Silverwrist Design Studios.  Portions created by Eric J. Bowersox are
 * Copyright (C) 2001 Eric J. Bowersox/Silverwrist Design Studios.  All Rights Reserved.
 * 
 * Contributor(s): 
 */
package com.silverwrist.venice.servlets;

import java.io.*;
import javax.servlet.*;
import javax.servlet.http.*;
import org.apache.log4j.*;
import com.silverwrist.util.StringUtil;
import com.silverwrist.venice.ValidationException;
import com.silverwrist.venice.core.*;
import com.silverwrist.venice.servlets.format.*;

public class SIGAdmin extends VeniceServlet
{
  /*--------------------------------------------------------------------------------
   * Static data members
   *--------------------------------------------------------------------------------
   */

  private static Category logger = Category.getInstance(SIGAdmin.class.getName());

  /*--------------------------------------------------------------------------------
   * Internal functions
   *--------------------------------------------------------------------------------
   */

  private SIGAdminTop makeSIGAdminTop() throws ServletException
  {
    final String desired_name = "SIGAdminTop";
    MenuPanelCache cache = MenuPanelCache.getMenuPanelCache(getServletContext());

    if (!(cache.isCached(desired_name)))
    { // create a template and save it off
      SIGAdminTop template = new SIGAdminTop();
      cache.saveTemplate(template);

    } // end if

    // return a new copy
    return (SIGAdminTop)(cache.getNewMenuPanel(desired_name));

  } // end makeSIGAdminTop

  private EditSIGProfileDialog makeEditSIGProfileDialog() throws ServletException
  {
    final String desired_name = "EditSIGProfileDialog";
    DialogCache cache = DialogCache.getDialogCache(getServletContext());

    if (!(cache.isCached(desired_name)))
    { // create a template and save it off
      EditSIGProfileDialog template = new EditSIGProfileDialog(getCountryList(),getLanguageList());
      cache.saveTemplate(template);

    } // end if

    // return a new copy
    return (EditSIGProfileDialog)(cache.getNewDialog(desired_name));

  } // end makeEditSIGProfileDialog

  /*--------------------------------------------------------------------------------
   * Overrides from class HttpServlet
   *--------------------------------------------------------------------------------
   */

  public String getServletInfo()
  {
    String rc = "SIGAdmin servlet - Administrative functions for SIGs\n"
              + "Part of the Venice Web Communities System\n";
    return rc;

  } // end getServletInfo

  /*--------------------------------------------------------------------------------
   * Overrides from class VeniceServlet
   *--------------------------------------------------------------------------------
   */

  protected VeniceContent doVeniceGet(HttpServletRequest request, VeniceEngine engine,
				      UserContext user, RenderData rdat)
    throws ServletException, IOException, VeniceServletResult
  {
    // get the SIG context
    SIGContext sig = getSIGParameter(request,user,true,"top");
    changeMenuSIG(request,sig);
    setMyLocation(request,"sigadmin?" + request.getQueryString());
    String on_error = "sigadmin?sig=" + sig.getSIGID();

    if (logger.isDebugEnabled())
      logger.debug("SIGAdmin/doGet operating on SIG \"" + sig.getName() + "\" (" + sig.getSIGID() + ")");

    // now decide what to do based on the "cmd" parameter
    String cmd = getStandardCommandParam(request);
    if (logger.isDebugEnabled())
      logger.debug("SIGAdmin/doGet command value = " + cmd);

    if (cmd.equals("P"))
    { // "P" = "Edit Profile"
      if (!(sig.canModifyProfile()))
      { // no access - sorry, dude
	logger.error("tried to call up SIG profile screen without access...naughty naughty!");
	return new ErrorBox("Unauthorized","You do not have access to modify this SIG's profile.",on_error);

      } // end if

      // construct the edit profile dialog and load it up for use
      EditSIGProfileDialog dlg = makeEditSIGProfileDialog();

      try
      { // load the values for this dialog
	dlg.setupDialog(engine,sig);

      } // end try
      catch (DataException de)
      { // we could not load the values because of a data exception
	logger.error("DB error loading SIG profile: " + de.getMessage(),de);
	return new ErrorBox("Database Error","Database error retrieving profile: " + de.getMessage(),on_error);

      } // end catch
      catch (AccessError ae)
      { // we don't have enough privilege
	logger.error("Access error loading SIG profile: " + ae.getMessage(),ae);
	return new ErrorBox("Unauthorized","You do not have access to modify this SIG's profile.",on_error);

      } // end catch

      return dlg;  // display me!

    } // end if ("P" command)

    if (cmd.equals("T"))
    { // "T" = "Set Category"
      if (!(sig.canModifyProfile()))
      { // no access - sorry man
	logger.error("tried to call up SIG category set screen without access...naughty naughty!");
	return new ErrorBox("Unauthorized","You do not have access to modify this SIG's profile.",on_error);

      } // end else

      // did they actually send a "set" parameter?
      String p = request.getParameter("set");
      if (!(StringUtil.isStringEmpty(p)))
      { // OK, we're setting the category ID...
	try
	{ // get the new category ID and set it
	  int catid = Integer.parseInt(p);
	  if (!(engine.isValidCategoryID(catid)))
	    throw new NumberFormatException();  // dump the category if it's not valid

	  // change the category ID
	  sig.setCategoryID(catid);

	} // end try
	catch (NumberFormatException nfe)
	{ // we got an invalid category value...
	  return new ErrorBox("Invalid Input","Invalid category ID passed to category browser.",on_error);

	} // end catch
	catch (DataException de)
	{ // unable to update the SIG properly
	  logger.error("DB error updating SIG: " + de.getMessage(),de);
	  return new ErrorBox("Database Error","Database error updating SIG: " + de.getMessage(),on_error);

	} // end catch
	catch (AccessError ae)
	{ // database access error - display an error box
	  logger.error("Access error updating SIG: " + ae.getMessage(),ae);
	  return new ErrorBox("Access Error",ae.getMessage(),on_error);

	} // end catch

	// now that it's set, go back to the admin menu
	throw new RedirectResult(on_error);

      } // end if (setting category ID)
      else
      { // we're browsing - try and figure out what to display
	try
	{ // get the "go" parameter to see what the current category ID is
	  p = request.getParameter("go");
	  int curr_catid;
	  if (StringUtil.isStringEmpty(p))
	    curr_catid = sig.getCategoryID();
	  else
	    curr_catid = Integer.parseInt(p);

	  if (!(engine.isValidCategoryID(curr_catid)))
	    throw new NumberFormatException();  // dump the category if it's not valid

	  // create the browser panel and let it rip
	  return new SIGCategoryBrowseData(user,sig,curr_catid);

	} // end try
	catch (NumberFormatException nfe)
	{ // we got an invalid category value...
	  return new ErrorBox("Invalid Input","Invalid category ID passed to category browser.",on_error);

	} // end catch
	catch (DataException de)
	{ // unable to get the categories properly
	  logger.error("DB error browsing categories: " + de.getMessage(),de);
	  return new ErrorBox("Database Error","Database error browsing categories: " + de.getMessage(),
			      on_error);

	} // end catch

      } // end else

    } // end if ("T" command)

    // all unknown requests get turned into menu display requests
    if (!(sig.canAdministerSIG()))
    { // no access - sorry buddy
      logger.error("tried to call up SIG admin menu without access...naughty naughty!");
      return new ErrorBox("Access Error","You do not have access to administer this SIG.",on_error);

    } // end if

    // create a SIG administration menu
    SIGAdminTop menu = makeSIGAdminTop();
    menu.setSIG(sig);
    return menu;

  } // end doVeniceGet

  protected VeniceContent doVenicePost(HttpServletRequest request, VeniceEngine engine,
				       UserContext user, RenderData rdat)
    throws ServletException, IOException, VeniceServletResult
  {
    // get the SIG context
    SIGContext sig = getSIGParameter(request,user,true,"top");
    changeMenuSIG(request,sig);
    String on_error = "sigadmin?sig=" + sig.getSIGID();
    setMyLocation(request,on_error);

    if (logger.isDebugEnabled())
      logger.debug("SIGAdmin/doPost operating on SIG \"" + sig.getName() + "\" (" + sig.getSIGID() + ")");

    // now decide what to do based on the "cmd" parameter
    String cmd = getStandardCommandParam(request);
    if (logger.isDebugEnabled())
      logger.debug("SIGAdmin/doPost command value = " + cmd);

    if (cmd.equals("P"))
    { // "P" = "Edit Profile"
      if (!(sig.canModifyProfile()))
      { // no access - sorry, dude
	logger.error("tried to call up SIG profile screen without access...naughty naughty!");
	return new ErrorBox("Unauthorized","You do not have access to modify this SIG's profile.",on_error);

      } // end if

      // construct the edit profile dialog and load it up for use
      EditSIGProfileDialog dlg = makeEditSIGProfileDialog();
      dlg.setupDialogBasic(engine,sig);

      if (dlg.isButtonClicked(request,"cancel"))
	throw new RedirectResult(on_error);  // go back - they canceled out

      if (dlg.isButtonClicked(request,"update"))
      { // begin updating the SIG's contents
	dlg.loadValues(request);  // do value loading now

	try
	{ // attempt to change the SIG profile now...
	  dlg.doDialog(sig);

	  // now jump back to the main menu
	  clearMenu(request);
	  throw new RedirectResult(on_error);

	} // end try
	catch (ValidationException ve)
	{ // simple validation exception
	  if (logger.isDebugEnabled())
	    logger.debug("validation failure: " + ve.getMessage());
	  dlg.resetOnError(sig,ve.getMessage() + " Please try again.");

	} // end catch
	catch (DataException de)
	{ // database error updating the SIG values
	  logger.error("DB error updating SIG: " + de.getMessage(),de);
	  return new ErrorBox("Database Error","Database error updating SIG: " + de.getMessage(),on_error);

	} // end catch
	catch (AccessError ae)
	{ // access error changing the SIG values
	  logger.error("Access error updating SIG: " + ae.getMessage(),ae);
	  return new ErrorBox("Access Error",ae.getMessage(),on_error);

	} // end catch

	return dlg;

      } // end if ("update" pressed)

      // what the hell was that button?
      logger.error("no known button click on SIGAdmin.doPost, cmd=P");
      return new ErrorBox("Internal Error","Unknown command button pressed",on_error);

    } // end if ("P" command)

    // on unknown command, redirect to the GET function
    return doVeniceGet(request,engine,user,rdat);

  } // end doVenicePost

} // end class SIGAdmin
